#ifndef _mod_dash_proxy_c
#define _mod_dash_proxy_c
/* -------------------------------------------------------------------- */
/*  			Amenesik Cloud Streaming                        */
/* -------------------------------------------------------------------- */
/*				DASH PROXY				*/
/* -------------------------------------------------------------------- */
/* (C) 2016,2018 by Iain James Marshall (Amenesik) <ijm667@hotmail.com> */
/* -------------------------------------------------------------------- */
/* The DASH PROXY is a Module for Apache 2.4 Web Server and allows media*/
/* files available at an origin server in DASH format to be fetched and */
/* cached for subsequent delivery to DASH client viewers.		*/
/* DASH media comprises manifest files MPD and segment files M4S.	*/
/* -------------------------------------------------------------------- */
/*				MODULE					*/
/* -------------------------------------------------------------------- */
/* The module definition file "dash_proxy.load" in the mods-available 	*/
/* directory of the apache configuration dirctory should contain the 	*/
/* following instruction:						*/
/*									*/
/* LoadModule dash_proxy_module /usr/lib/apache2/modules/mod_dash_proxy.so*/
/*									*/
/* and can be enabled by launching the script:				*/
/*									*/
/*	a2enmod dash_proxy						*/
/*	service apache2 restart						*/
/*									*/
/* -------------------------------------------------------------------- */
/*			     CONFIGURATION				*/
/* -------------------------------------------------------------------- */
/* DASH Proxy expects the following directives in Apache Configuration :*/
/*									*/
/*	<Location "/stream">						*/
/*        	SetHandler dash-proxy-handler				*/
/*	</Location>							*/
/*									*/
/*	AddHandler dash-proxy-handler m3u8				*/
/*	AddHandler dash-proxy-handler ts				*/
/*									*/
/*	DashProxyEnabled On						*/
/*	DashProxyPort "80"						*/
/*	DashProxyTtl  "5"						*/
/*	DashProxyDebug "0"						*/
/*	DashProxyRetry "0"						*/
/*	DashProxyRetries "10"						*/
/*	DashProxyAgent "dash-proxy/1.0a"				*/
/*	DashProxyPath "Local path prefix to be removed for remote fetch"*/
/*	DashProxyName "name of remote server"				*/
/*	DashProxyUser "remote server auth user [none|username]"		*/
/*	DashProxyPass "remote server auth pass [none|password]"		*/
/*	DashProxyAuth "remote server auth scheme [none|basic|digest|etc]"*/
/*									*/
/*	DashProxyHost "url of the origin server"			*/
/*									*/
/* -------------------------------------------------------------------- */
#include <stdio.h>
#include "apr_hash.h"
#include "ap_config.h"
#include "ap_provider.h"
#include "httpd.h"
#include "http_core.h"
#include "http_config.h"
#include "http_log.h"
#include "http_protocol.h"
#include "http_request.h"
#include "apr_tables.h"
#include "apr_file_io.h"
#include "apr_time.h"
#include "apr_network_io.h"
#include "apr_strings.h"

/*	-----------	*/
/*	definitions 	*/
/*	-----------	*/
#define	_DASH_INFO		1
#define	_DEBUG_DASH_CHUNKS	1
#define	_DASH_PROXY_TRANSFER_BUFFER_SIZE	512000

#define	isdebug DashProxy.debug

#ifndef	private
#define	private static
#endif

#ifndef	public
#define	public
#endif

#ifndef	OK
#define	OK 200
#endif

#ifndef	HTTP_NOT_FOUND
#define	HTTP_NOT_FOUND 404
#endif

#ifndef	INCORRECT_METHOD
#define	INCORRECT_METHOD 405
#endif

#ifndef	HTTP_BAD_REQUEST
#define	HTTP_BAD_REQUEST 400
#endif

#define	dash_log_message(r,s,m,x) ap_log_error_(__FILE__,__LINE__,999,APLOG_NOTICE,0,r->server,"%s %s",m,x)

/*	-----------------------------------------------	*/
/*	d a s h _ p r o x y _ c o n f i g u r a t i o n	*/
/*	-----------------------------------------------	*/
struct	dash_proxy_configuration
{
	const char *	host;
	const char *	path;
	const char *	name;
	const char *	user;
	const char *	pass;
	const char *	auth;
	const char *	agent;
	int		ttl;
	int		port;
	int		retries;
	int		retry;
	int		enabled;
	int		debug;
};

/*	-----------------------------------	*/
/*	d a s h _ p r o x y _ c o n t e x t	*/
/*	-----------------------------------	*/
struct	dash_proxy_context
{
	request_rec *	request;
	apr_file_t *	handle;
	apr_socket_t * 	socket;
	int		bufsize;
	int		consumed;
	apr_size_t 	bytes;
	char		buffer[_DASH_PROXY_TRANSFER_BUFFER_SIZE];
};

struct  dash_proxy_header
{
	char *  name;
	char *  value;
        
};

/*	----------	*/
/*	prototypes	*/
/*	----------	*/

private	int	dash_cache_manifest(char * filename, request_rec *r);
private	int	dash_cache_segment(char * filename, request_rec *r);
private	int	dash_cached_object(char * filename, request_rec *r);

private const char * dash_proxy_set_enabled(cmd_parms *cmd, void *cfg, const char *arg);
private const char * dash_proxy_set_debug(cmd_parms *cmd, void *cfg, const char *arg);
private const char * dash_proxy_set_host(cmd_parms *cmd, void *cfg, const char *arg);
private const char * dash_proxy_set_path(cmd_parms *cmd, void *cfg, const char *arg);
private const char * dash_proxy_set_name(cmd_parms *cmd, void *cfg, const char *arg);
private const char * dash_proxy_set_user(cmd_parms *cmd, void *cfg, const char *arg);
private const char * dash_proxy_set_auth(cmd_parms *cmd, void *cfg, const char *arg);
private const char * dash_proxy_set_pass(cmd_parms *cmd, void *cfg, const char *arg);
private const char * dash_proxy_set_agent(cmd_parms *cmd, void *cfg, const char *arg);
private const char * dash_proxy_set_port(cmd_parms *cmd, void *cfg, const char *arg);
private const char * dash_proxy_set_ttl(cmd_parms *cmd, void *cfg, const char *arg);
private const char * dash_proxy_set_retry(cmd_parms *cmd, void *cfg, const char *arg);
private const char * dash_proxy_set_retries(cmd_parms *cmd, void *cfg, const char *arg);

private	void 	register_hooks(apr_pool_t *pool);

/*	-----------------	*/
/*	global structures	*/
/*	-----------------	*/

private	const command_rec        dash_proxy_directives[] =
{
    AP_INIT_TAKE1("DashProxyEnabled", dash_proxy_set_enabled, NULL, RSRC_CONF, "Enable or disable DASH Proxy Module"),
    AP_INIT_TAKE1("DashProxyDebug", dash_proxy_set_debug, NULL, RSRC_CONF, "The Origen Server Retry Sleep "),
    AP_INIT_TAKE1("DashProxyHost", dash_proxy_set_host, NULL, RSRC_CONF, "The Origin Server Host"),
    AP_INIT_TAKE1("DashProxyPath", dash_proxy_set_path, NULL, RSRC_CONF, "The Origin Server Host"),
    AP_INIT_TAKE1("DashProxyName", dash_proxy_set_name, NULL, RSRC_CONF, "The Origin Server Host"),
    AP_INIT_TAKE1("DashProxyUser", dash_proxy_set_user, NULL, RSRC_CONF, "The Origin Server Host"),
    AP_INIT_TAKE1("DashProxyAuth", dash_proxy_set_auth, NULL, RSRC_CONF, "The Origin Server Host"),
    AP_INIT_TAKE1("DashProxyPass", dash_proxy_set_pass, NULL, RSRC_CONF, "The Origin Server Host"),
    AP_INIT_TAKE1("DashProxyAgent", dash_proxy_set_agent, NULL, RSRC_CONF, "The DashProxy Agent Name"),
    AP_INIT_TAKE1("DashProxyPort", dash_proxy_set_port, NULL, RSRC_CONF, "The Origin Server Port"),
    AP_INIT_TAKE1("DashProxyTtl", dash_proxy_set_ttl, NULL, RSRC_CONF, "The Origin Server Port"),
    AP_INIT_TAKE1("DashProxyRetry", dash_proxy_set_retry, NULL, RSRC_CONF, "The Origen Server Retry Sleep "),
    AP_INIT_TAKE1("DashProxyRetries", dash_proxy_set_retries, NULL, RSRC_CONF, "The Origin Server Retry Count"),
    { NULL }
};

module AP_MODULE_DECLARE_DATA   dash_proxy_module =
{
    STANDARD20_MODULE_STUFF,
    NULL,
    NULL,
    NULL,
    NULL,
    dash_proxy_directives,
    register_hooks   
};

/*	----------------------------------	*/
/*	DASH PROXY CONFIGURATION FUNCTIONS	*/
/*	----------------------------------	*/

private	struct dash_proxy_configuration DashProxy = 
{
	(char *) 0,			/* remote stream origin host */
	(char *) "/var/www/html",	/* locate path prefix	     */
	(char *) "DASH PROXY",		/* remote stream origin name */
	(char *) "none",		/* remote stream origin user */
	(char *) "none",		/* remote stream origin pass */
	(char *) "none",		/* remote stream origin auth */
	(char *)"dash-proxy/1.0a",	/* local server/user-agent name  */
	80,				/* standard HTTP port */
	5,				/* standard HTTP TTL */
	10,				/* retry counter*/
	0,				/* time between retries*/
	0,				/* not enabled */
	0				/* debug not active */
};
	
/*	----------------------	*/
/*	dash_proxy_set_enabled	*/
/*	----------------------	*/
private const char * dash_proxy_set_enabled(cmd_parms *cmd, void *cfg, const char *arg)
{
	if(!strcasecmp(arg, "on")) 
		DashProxy.enabled = 1;
	else 	DashProxy.enabled = 0;
	return( NULL );
}

/*	-------------------	*/
/*	dash_proxy_set_host	*/
/*	-------------------	*/
private const char * dash_proxy_set_host(cmd_parms *cmd, void *cfg, const char *arg)
{
	DashProxy.host = arg;
	return( NULL );
}

/*	-------------------	*/
/*	dash_proxy_set_path	*/
/*	-------------------	*/
private const char * dash_proxy_set_path(cmd_parms *cmd, void *cfg, const char *arg)
{
	DashProxy.path = arg;
	return( NULL );
}

/*	-------------------	*/
/*	dash_proxy_set_name	*/
/*	-------------------	*/
private const char * dash_proxy_set_name(cmd_parms *cmd, void *cfg, const char *arg)
{
	DashProxy.name = arg;
	return( NULL );
}

/*	-------------------	*/
/*	dash_proxy_set_user	*/
/*	-------------------	*/
private const char * dash_proxy_set_user(cmd_parms *cmd, void *cfg, const char *arg)
{
	DashProxy.user = arg;
	return( NULL );
}

/*	-------------------	*/
/*	dash_proxy_set_pass	*/
/*	-------------------	*/
private const char * dash_proxy_set_pass(cmd_parms *cmd, void *cfg, const char *arg)
{
	DashProxy.pass = arg;
	return( NULL );
}

/*	-------------------	*/
/*	dash_proxy_set_auth	*/
/*	-------------------	*/
private const char * dash_proxy_set_auth(cmd_parms *cmd, void *cfg, const char *arg)
{
	DashProxy.auth = arg;
	return( NULL );
}
/*	--------------------	*/
/*	dash_proxy_set_agent	*/
/*	--------------------	*/
private const char * dash_proxy_set_agent(cmd_parms *cmd, void *cfg, const char *arg)
{
	DashProxy.agent = arg;
	return( NULL );
}

/*	-------------------	*/
/*	dash_proxy_set_port	*/
/*	-------------------	*/
private const char * dash_proxy_set_port(cmd_parms *cmd, void *cfg, const char *arg)
{
	DashProxy.port = atoi(arg);
	return( NULL );
}

/*	------------------	*/
/*	dash_proxy_set_TTL 	*/
/*	------------------	*/
private const char * dash_proxy_set_ttl(cmd_parms *cmd, void *cfg, const char *arg)
{
	DashProxy.ttl = atoi(arg);
	return( NULL );
}

/*	--------------------	*/
/*	dash_proxy_set_debug	*/
/*	--------------------	*/
private const char * dash_proxy_set_debug(cmd_parms *cmd, void *cfg, const char *arg)
{
	DashProxy.debug = atoi(arg);
	return( NULL );
}

/*	--------------------	*/
/*	dash_proxy_set_retry	*/
/*	--------------------	*/
private const char * dash_proxy_set_retry(cmd_parms *cmd, void *cfg, const char *arg)
{
	DashProxy.retry = atoi(arg);
	return( NULL );
}

/*	----------------------	*/
/*	dash_proxy_set_retries	*/
/*	----------------------	*/
private const char * dash_proxy_set_retries(cmd_parms *cmd, void *cfg, const char *arg)
{
	DashProxy.retries = atoi(arg);
	return( NULL );
}

/*	--------------------------------	*/
/*	HTTP CLIENT and SOCKET FUNCTIONS	*/
/*	--------------------------------	*/

/*	---------------------	*/
/*	dash_proxy_disconnect	*/
/*	---------------------	*/
private apr_socket_t * dash_proxy_disconnect( struct dash_proxy_context * context )
{
	apr_status_t status;

	if ((status = apr_socket_close(context->socket)) != APR_SUCCESS)
                                        
		return( (apr_socket_t *) 0 );
                                
	else    return( (apr_socket_t *) 0 );

}

/*	---------------------------------	*/
/*	d a s h _ p r o x y _ s o c k e t	*/
/*	---------------------------------	*/
private	int	dash_proxy_socket(apr_pool_t *pool, apr_socket_t **socket, int *family)
{
	apr_status_t status;
	apr_sockaddr_t * sockadr;
    
	if (( status = apr_socket_create(socket, *family, SOCK_STREAM, 0, pool)) != APR_SUCCESS)
		return(0);

	if ( *family == APR_UNSPEC ) 
	{
		if (( status = apr_socket_addr_get(&sockadr, APR_LOCAL, *socket)) != APR_SUCCESS) 
			return(0);
		else	*family = sockadr->family;
	}
	return(1);
}

/*	------------------	*/
/*	dash_proxy_receive	*/
/*	------------------	*/
private	int	dash_proxy_receive( struct dash_proxy_context * context )
{
	context->consumed = 0;
	context->bytes = context->bufsize;
	apr_socket_recv(context->socket, context->buffer, &context->bytes);
	return(context->bytes);
}

/*	------------------	*/
/*	dash_proxy_getbyte	*/
/*	------------------	*/
private	int	dash_proxy_getbyte( struct dash_proxy_context *context )
{
	if ( context->bytes <= context->consumed )
	{
		if (!( context->bytes = dash_proxy_receive( context ) ))
			return( -1 );
		else	context->consumed= 0;
	}
	return(( context->buffer[ context->consumed++ ] & 0x00FF ));
}

/*	------------------	*/
/*	dash_proxy_getline	*/
/*	------------------	*/
private	int	dash_proxy_getline( struct dash_proxy_context * context, char *buffer, int buflen )
{
	int	bytes=0;
	int	c;
	while ((c = dash_proxy_getbyte( context )) > 0 )
	{
		if ( c =='\r' )
			continue;
		else if ( c =='\n' )
			break;
		else
		{
			*(buffer+bytes) = c;
			bytes++;
			if ( bytes > (buflen-2) )
				return( 0 );
			else	continue;
		}
	}
	*(buffer+bytes) = 0;
	return( bytes );
}

/*	----------------	*/
/*	dash_proxy_chunk	*/
/*	----------------	*/
private	void	dash_proxy_chunk(request_rec *r,  apr_size_t bytes)
{
	char	buffer[256];
	sprintf(buffer,"%x",(unsigned int) bytes);
	if ( isdebug > 1 ) dash_log_message(r, 0,"Response Chunk Size",buffer);
	if (!( bytes ))
		strcat(buffer,"\r\n\r\n");
	else	strcat(buffer,"\r\n");
	ap_rwrite(buffer,strlen(buffer),r);
	return;
}

/*	-----------------------------------	*/
/*	d a s h _ p r o x y _ c o n t e n t 	*/
/*	-----------------------------------	*/
/*	arrival here after reading header	*/
/*	with data in the buffer. Will be	*/
/*	fetched using content length info	*/
/*	from the origin server and stored	*/
/*	into the cache file while sending	*/
/*	back to the client at same time		*/
/*	using chunked TE.			*/
/*	-----------------------------------	*/
private	int	dash_proxy_content(  struct dash_proxy_context * context, int content_length )
{
	apr_size_t transfer=0;
	char	buffer[256];

	sprintf(buffer,"%u",content_length);
	/* ---------------------------------------------------- */
	/* signal chunked transfer encoding for result delivery */
	/* ---------------------------------------------------- */
	if ( isdebug > 1 ) dash_log_message(context->request, 0,"Response Content-Length", buffer );
	apr_table_setn(context->request->headers_out, "Transfer-Encoding", "chunked");

	while (content_length > 0)
	{
		/* ------------------------------------------ */
		/* attempt to fetch from remote origin server */
		/* ------------------------------------------ */
		if ((transfer = ( context->bytes - context->consumed )) > content_length)
			transfer = content_length;

		if ( transfer > 0 )
		{
			/* ------------------------ */
			/* write this block to disk */
			/* ------------------------ */
			if ( apr_file_write( context->handle, (context->buffer+context->consumed), &transfer ) != APR_SUCCESS )
			{
				dash_log_message(context->request, HTTP_INTERNAL_SERVER_ERROR,"failure writing", "cache file" );
				return( HTTP_INTERNAL_SERVER_ERROR );
			}

			/* ------------------------------- */
			/* write this chunk to the network */
			/* ------------------------------- */
			dash_proxy_chunk(context->request,transfer);
			if ( ap_rwrite( (context->buffer+context->consumed), transfer, context->request) != transfer )
			{
				dash_log_message(context->request, HTTP_INTERNAL_SERVER_ERROR,"failure writing", "client socket" );
				return( HTTP_INTERNAL_SERVER_ERROR );
			}
			else	
			{
				ap_rwrite( "\r\n", 2, context->request);
				content_length -= transfer;
				context->consumed += transfer;
			}
		}
		/* --------------------------- */
		/* receive a new block of data */
		/* --------------------------- */
		else if (!( dash_proxy_receive( context ) ))
			return( HTTP_BAD_REQUEST );
		else	continue;
	}
	dash_proxy_chunk(context->request,0);
	if ( isdebug > 1 ) dash_log_message(context->request, 0,"content_length transfer", "completed" );
	return( OK );
}

/*	--------------------	*/
/*	dash_proxy_check_eol	*/
/*	--------------------	*/
private	int	dash_proxy_consume_eol( struct dash_proxy_context * context )
{
	int	c;
	while ((c = dash_proxy_getbyte( context )) > 0 )
	{
		if ( c == '\r' )
			continue;
		else 	break;
	}
	return( c );
}
	
/*	---------------------	*/
/*	dash_proxy_chunk_size	*/
/*	---------------------	*/
private	int	dash_proxy_chunk_size( struct dash_proxy_context * context )
{
	int	c;
	int	chunksize=0;
	int	comment=0;
	while ((c = dash_proxy_getbyte( context )) > 0 )
	{
		if ( c == '\r' )
			continue;
		else if ( c == ' ' )
			continue;
		else if ( c == '\t' )
			continue;
		else if ( c == '\n' )
			break;
		else if ( comment )
			continue;
		else if (( c >= '0' ) && (c <= '9'))
			chunksize = ((chunksize * 16) + (c - '0'));
		else if (( c >= 'a' ) && (c <= 'f'))
			chunksize = ((chunksize * 16) + ((c - 'a') + 10));
		else if (( c >= 'A' ) && (c <= 'F'))
			chunksize = ((chunksize * 16) + ((c - 'A') + 10));
		else if ( c == ';' )
			comment = 1;
		else	return( 0 );
	}
	return( chunksize );		
}

/*	-----------------------------------	*/
/*	d a s h _ p r o x y _ c h u n k e d	*/
/*	-----------------------------------	*/
/*	arrival here after reading header	*/
/*	with data in the buffer. Will be	*/
/*	fetched using chunked delivery 		*/
/*	from the origin server and stored	*/
/*	into the cache file while sending	*/
/*	back to the client at same time		*/
/*	using chunked TE.			*/
/*	-----------------------------------	*/
private	int	dash_proxy_chunked(  struct dash_proxy_context * context )
{
	apr_size_t chunksize;
	apr_size_t transfer;
	int	total=0;
	int	c;
	char buffer[2048];

	/* ---------------------------------------------------- */
	/* signal chunked transfer encoding for result delivery */
	/* ---------------------------------------------------- */
	if ( isdebug > 1 ) dash_log_message(context->request, OK,"Response Transfer-Encoding", "chunked");
	apr_table_setn(context->request->headers_out, "Transfer-Encoding", "chunked");

	while ((chunksize = dash_proxy_chunk_size( context )) > 0)
	{
		total += chunksize;
		while ( chunksize )
		{
			/* --------------------------------------------------------- */
			/* calculate possible transfer from remaining data in buffer */
			/* --------------------------------------------------------- */
			if ((transfer = (context->bytes - context->consumed)) > chunksize)
				transfer = chunksize;

			if ( transfer > 0 )
			{

				/* ------------------------ */
				/* write this block to disk */
				/* ------------------------ */
				if ( apr_file_write( context->handle, (context->buffer+context->consumed), &transfer ) != APR_SUCCESS )
				{
					dash_log_message(context->request, HTTP_INTERNAL_SERVER_ERROR,"failure writing", "cache file" );
					return( HTTP_INTERNAL_SERVER_ERROR );
				}

				/* ------------------------------- */
				/* write this chunk to the network */
				/* ------------------------------- */
				dash_proxy_chunk(context->request,transfer);
				if ( ap_rwrite( (context->buffer+context->consumed), transfer, context->request) != transfer )
				{
					dash_log_message(context->request, HTTP_INTERNAL_SERVER_ERROR,"failure writing", "client socket" );
					return( HTTP_INTERNAL_SERVER_ERROR );
				}
				else	
				{
					/* ------------------------- */
					/* terminate this sent chunk */
					/* ------------------------- */
					ap_rwrite( "\r\n", 2, context->request);
					chunksize -= transfer;
					context->consumed += transfer;
				}
			}
			else if ( dash_proxy_receive( context ) <= 0 )
			{
				dash_log_message(context->request, 0,"unexpected EOF", "in source stream" );
				return( HTTP_INTERNAL_SERVER_ERROR );
			}
		}
		/* ------------------------------ */
		/* remove end of line after chunk */
		/* ------------------------------ */
		if ((c =  dash_proxy_consume_eol( context )) != '\n' )
		{
			dash_log_message(context->request, 0,"expected EOL", "in source stream" );
			return( HTTP_INTERNAL_SERVER_ERROR );
		}
	}

	/* -------------------------------------------- */
	/* remove trailing headers and final empty line */
	/* -------------------------------------------- */
	while ( dash_proxy_getline( context, buffer, 2048) > 0 );

	/* -------------------------- */
	/* signal end of chunk stream */
	/* -------------------------- */
	dash_proxy_chunk(context->request,0);

	if ( isdebug > 1 ) dash_log_message(context->request, 0,"chunked transfer", "completed" );
	return( OK );
}

/*	--------------------------	*/
/*	dash_proxy_response_status	*/
/*	--------------------------	*/
private int     dash_proxy_response_status(  struct dash_proxy_context * context, char * buffer )
{
	if ( isdebug ) dash_log_message( context->request, 0, "Origin Server HTTP Response",buffer);

	/* ---------------------- */
	/* step over HTTP version */
	/* ---------------------- */
	while (( *buffer != 0 ) && ( *buffer != ' ' )) buffer++;

	/* ---------------------- */
	/* return the status code */
	/* ---------------------- */
	if (!( *buffer )) 
		return( 500 );
	else	return( atoi( (buffer+1) ));

}

/*	-----------------------		*/
/*	dash_proxy_parse_header		*/
/*	-----------------------		*/
private	int	dash_proxy_parse_header( char * buffer, struct  dash_proxy_header * header )
{
	header->name = buffer;
	while (( *buffer != 0 ) && ( *buffer != ':' )) buffer++;
	if ( *buffer == ':' )
	{
		*(buffer++) = 0;
		while ((*buffer != 0) && (*buffer == ' ')) 
			buffer++;
	}
	header->value = buffer;
	return(1);
}

/*	-------------------------------------	*/
/*	d a s h _ p r o x y _ r e s p o n s e 	*/
/*	-------------------------------------	*/
private	int	dash_proxy_response(  struct dash_proxy_context * context )
{
	int	content_length=0;
	int	chunked=0;
	int	status;
	char	buffer[16000];
	struct	dash_proxy_header header = { (char*) 0,(char *) 0 };

	/* ------------------------------------------ */
	/* collect and process the HTTP response line */
	/* ------------------------------------------ */
	if (!( dash_proxy_receive( context )))
	{
		dash_log_message(context->request, 0,"expected","response from server" );
		return( HTTP_BAD_REQUEST );
	}

	else if (!( dash_proxy_getline( context, buffer, 16000 )))
	{
		dash_log_message(context->request, 0,"expected", "origin server response line" );
		return( HTTP_BAD_REQUEST );
	}

	/* ------------------------------------ */
	/* check the response status code is OK */
	/* ------------------------------------ */
	else if ((status = dash_proxy_response_status( context, buffer )) != 200 )
	{
		dash_log_message(context->request, status ,"expected", "origin server 200" );
		return( HTTP_BAD_REQUEST );
	}

	/* ------------------------------------------ */
	/* collect and process the HTTP header fields */
	/* ------------------------------------------ */
	while ( dash_proxy_getline( context, buffer, 16000 ) > 0 )
	{
		/* ---------------- */
		/* collect a header */
		/* ---------------- */
		if (!( dash_proxy_parse_header(apr_pstrdup(context->request->pool, buffer), &header ) ))
		{
			dash_log_message(context->request, HTTP_BAD_REQUEST,"expected", "origin server header" );
			return( HTTP_BAD_REQUEST );
		}

		if ( isdebug > 1 ) dash_log_message(context->request, 0, header.name, header.value );

		/*----------------------------- */
		/* detect content length header */
		/*----------------------------- */
		if (!( strcasecmp( header.name, "content-length" ) ))
			content_length = atoi( header.value );

		/* --------------------------------------- */
		/* detect chunked transfer encoding header */
		/* --------------------------------------- */
		else if (!( strcasecmp( header.name, "transfer-encoding" ) ))
			if (!( strcasecmp( header.value,"chunked" ) ))
				chunked = 1;
	}

	/* ------------------------------------------ */
	/* collect and process the HTTP response body */
	/* ------------------------------------------ */
	if ( chunked )
		return( dash_proxy_chunked( context ) );
	else if ( content_length )
		return( dash_proxy_content( context, content_length ) );
	else
	{
		dash_log_message(context->request, HTTP_NOT_FOUND,"expected", "content-length or chunked" );
		return( HTTP_BAD_REQUEST );
	}
}

/*	-----------------------------------	*/
/*	d a s h _ p r o x y _ r e q u e s t	*/
/*	-----------------------------------	*/
private	int	dash_proxy_request(  struct dash_proxy_context * context, char * method, const char* host, const char * object )
{
	/* --------------------------------------------------- */ 
	/* prepare the HTTP request line and essential headers */
	/* --------------------------------------------------- */ 
	sprintf(context->buffer,"%s %s HTTP/1.1\r\nHost: %s\r\nUser-Agent: %s\r\n\r\n",
		method, object, host, DashProxy.agent);

	/* ------------------------------------------ */
	/* send the HTTP request to the origin server */
	/* ------------------------------------------ */
	context->bytes = strlen( context->buffer );	
	apr_socket_send(context->socket, context->buffer, &context->bytes);
	
	return(1);
}

/*	------------------	*/
/*	dash_proxy_connect	*/
/*	------------------	*/
private apr_socket_t * dash_proxy_connect( struct dash_proxy_context * context, const char *host, const char *object )
{
	apr_status_t 	status;
	char		buffer[256];
	apr_sockaddr_t *dsa=(apr_sockaddr_t *)0;
	int		family=APR_INET;
	int		retries=DashProxy.retries;
	int		retry=DashProxy.retry;

	/* ----------------------------------- */
	/* resolve the destination address now */
	/* ----------------------------------- */
    	if (( status = apr_sockaddr_info_get(&dsa, host, family, DashProxy.port, 0, context->request->pool)) != APR_SUCCESS) 
	{
		dash_log_message(context->request, OK,"Cannot resolve host", host);
		return( (apr_socket_t *) 0 );
	}

	/* ----------------------------- */
	/* attempt to connect the socket */
	/* ----------------------------- */
	while (retries--) 
	{
		/* ---------------------------------------- */
		/* attempt to create and configure a socket */
		/* ---------------------------------------- */
		if (!( dash_proxy_socket( context->request->pool, &context->socket, &family ) ))
		{
			dash_log_message(context->request, OK,"Socket Creation Failure", host);
			return( (apr_socket_t *) 0 );
		}
		else
		{
			if ( isdebug ) dash_log_message(context->request, OK,"Connecting to host ", host);
        		status = apr_socket_connect(context->socket, dsa);
        		if (APR_STATUS_IS_ECONNREFUSED(status))
			{
				dash_log_message(context->request, OK,"Connection refused by ", host);
            			if ((status = apr_socket_close(context->socket)) != APR_SUCCESS) 
					return( (apr_socket_t *) 0 );
				else	context->socket = (apr_socket_t *) 0;

				if ( retries )
				{
					if ( retry )
            					apr_sleep(retry);
            				retry *= 2;
				}
				else	
				{
					break;
				}
			}
			else	break;
		}
	}

	/* --------------------------- */
	/* test for connection failure */
	/* --------------------------- */
	if (!( context->socket )) 
		return( context->socket );

	/* --------------------- */
	/* send the HTTP request */
	/* --------------------- */
	if ( isdebug > 1 ) dash_log_message(context->request, OK,"Sending Request for ", object);

	if (!( dash_proxy_request( context, "GET", host,  object ) ))
		return( dash_proxy_disconnect( context ) );

	/* ---------------------- */
	/* recv the HTTP response */
	/* ---------------------- */
	else if ((status = dash_proxy_response( context )) != OK )
	{
		sprintf(buffer,"%u",(unsigned int) status);
		if ( isdebug > 1 ) dash_log_message(context->request, OK,"Received Response ", buffer);
		return( dash_proxy_disconnect( context ) );
	}

	/* ----------------- */
	/* return the socket */
	/* ----------------- */
	else	return( context->socket );
}

/*	---------------------------------	*/
/*	 d a s h  _ p r o x y _ f e t c h  	*/
/*	---------------------------------	*/
/*	fetches data from the remote end	*/
/*	point origin server.			*/
/*	If the socket is NULL then a new	*/
/*	connection willbe established.		*/
/*	Zero data returned means end of		*/
/*	datai stream and socket will be 	*/
/*	closed.					*/
/*	---------------------------------	*/
private	int	dash_proxy_fetch( struct dash_proxy_context * context, const char *host, const char *object )
{
	int	bytes;

	/* ------------------------------------------------------------ */
	/* attempt to connect to remote origin server if socket is NULL */
	/* ------------------------------------------------------------ */
	if (!( dash_proxy_connect( context, host, object )))
		return( HTTP_NOT_FOUND  );
	else
	{
		/* operation has been performed */
		/* ---------------------------- */
		dash_proxy_disconnect( context );
		return( OK );
	}
}

/*	-------------------	*/
/*	 dash_cache_object	*/
/*	-------------------	*/
private	int	dash_cache_object(char * filename, request_rec *r)
{
	/* --------------------------------- */
	/* prepare the context for the fetch */
	/* --------------------------------- */
	struct	dash_proxy_context context = 
	{
		r,
		(apr_file_t *) 0,
		(apr_socket_t *) 0,
		_DASH_PROXY_TRANSFER_BUFFER_SIZE,
		0,0
	};

	/* ------------------------------------------------ */
	/* create a new cache file for storage of this item */
	/* ------------------------------------------------ */
	if ( apr_file_open( &context.handle, filename, APR_FOPEN_WRITE|APR_FOPEN_CREATE, APR_OS_DEFAULT, r->pool ) != APR_SUCCESS)
	{
		dash_log_message(r, HTTP_NOT_FOUND,"cannot create", filename );
		return( HTTP_NOT_FOUND );
	}

	/* -------------------------------------------------- */
	/* fetch, store and transfer the object to the client */
	/* -------------------------------------------------- */
	else if ( dash_proxy_fetch( &context, DashProxy.host, r->filename+strlen(DashProxy.path)) != OK )
	{
		apr_file_close( context.handle );
		dash_log_message(r, HTTP_NOT_FOUND,DashProxy.host, r->filename+strlen(DashProxy.path) );
		return( HTTP_NOT_FOUND );
	}
	else
	{
		apr_file_close( context.handle );
		return( OK );
	}
}

/*	------------------	*/
/*	dash_cache_segment 	*/
/*	------------------	*/
private	int	dash_cache_segment(char * filename, request_rec *r)
{
	return( dash_cache_object( filename, r ) );
}

/*	-------------------	*/
/*	dash_cache_manifest	*/
/*	-------------------	*/
private	int	dash_cache_manifest(char * filename, request_rec *r)
{
	return( dash_cache_object( filename, r ) );
}

/*	-------------------	*/
/*	dash_cached_segment	*/
/*	-------------------	*/
private	int	dash_cached_object(char * filename, request_rec *r)
{
	apr_file_t * handle;
	apr_size_t bytes=0;
	int	status;
	char	buffer[_DASH_PROXY_TRANSFER_BUFFER_SIZE];
	char	debuger[2048];
	/* ------------------------ */
	/* attempt to open the file */
	/* ------------------------ */
	if ( isdebug ) dash_log_message(r, 0,"Cached Object", filename);

	if ((status = apr_file_open(&handle, filename, APR_FOPEN_READ, APR_OS_DEFAULT, r->pool)) != APR_SUCCESS)
	{
		dash_log_message(r, HTTP_NOT_FOUND,"cannot open", filename );
		return( HTTP_NOT_FOUND );
	}
	else
	{
		if ( isdebug > 1 ) dash_log_message(r, 0,"Response Transfer-Encoding", "chunked");
		apr_table_setn(r->headers_out, "Transfer-Encoding", "chunked");
		while (1)
		{
			/* -------------------- */
			/* read a block of data */
			/* -------------------- */
			bytes = _DASH_PROXY_TRANSFER_BUFFER_SIZE;
            		if ( apr_file_read(handle, buffer, &bytes) != APR_SUCCESS )
			{
				if ( isdebug > 1 ) dash_log_message(r, 0,"apr_file_read", "not success");
				break;
			}
			else if (!( bytes ))
			{
				if ( isdebug > 1 ) dash_log_message(r, 0,"apr_file_read", "not success");
				break;
			}
			else
			{
				/* --------------------------- */
				/* send the data to the client */
				/* --------------------------- */
				if ( isdebug > 1 )
				{
					sprintf(debuger,"read d%u(x%x) bytes",(unsigned int)bytes,(unsigned int)bytes);
					dash_log_message(r, 0,"apr_file_read", debuger);
				}
				/* ----------------- */
				/* signal chunk size */
				/* ----------------- */
				dash_proxy_chunk(r, bytes);

				/* ---------------------- */
				/* write chunk data block */
				/* ---------------------- */
				if ( ap_rwrite( buffer, bytes, r) != bytes )
				{
					sprintf(debuger,"failure not %u bytes",(unsigned int)bytes);
					dash_log_message(r, 0,"ap_rwrite", debuger);
					return( HTTP_INTERNAL_SERVER_ERROR );
				}
				ap_rwrite( "\r\n", 2, r );
			}
		}
		/* ---------------------- */
		/* terminate chunk stream */
		/* ---------------------- */
		dash_proxy_chunk(r,0);

		apr_file_close( handle );

		if ( isdebug > 1 ) dash_log_message(r, 0,"Cached Object delivery", "completed");
		return( OK );
	}
}

/*      ----------------------------------------        */
/*         d a s h _ p r o x y _ m a n i f e s t        */
/*      ----------------------------------------        */
private int dash_proxy_manifest(request_rec *r)
{
        apr_finfo_t finfo;
        apr_time_t now;
        int     status;
        int     exists;
        char * target=r->filename;
        char    buffer[1024];
        ap_set_content_type(r, "application/x-mpegURL");
        /* retrieve the file statistics */
        if ((status = apr_stat(&finfo, target, APR_FINFO_MIN, r->pool)) != APR_SUCCESS)
                return( dash_cache_manifest( target, r ) );
        /* ensure file exists */
        else if (!( exists = ((finfo.filetype != APR_NOFILE) &&  !(finfo.filetype & APR_DIR)) ))
                return( dash_cache_manifest( target, r ) );
        /* check for file expired */
        else if (( apr_time_sec(finfo.mtime) + DashProxy.ttl ) > (now=apr_time_sec(apr_time_now())) )
        {
                if ( isdebug )
                {
                        sprintf(buffer,"FILETIME(%u) + TTL(%u) >  NOW(%u)",(unsigned int)apr_time_sec(finfo.mtime),(unsigned int)DashProxy.ttl,(unsigned int)now);
                        dash_log_message(r,200,"Reusing File : ", buffer);
                }
                return( dash_cached_object( target, r ) );
        }
        /* fetch a new copy */
        else
        {
                if ( isdebug )
                {
                        sprintf(buffer,"FILETIME(%u) + TTL(%u) <= NOW(%u)",(unsigned int)apr_time_sec(finfo.atime),(unsigned int)DashProxy.ttl,(unsigned int)now);
                        dash_log_message(r,200,"Refresh File : ", buffer);
                }
                return( dash_cache_manifest( target, r ) );
        }
}

/*      ----------------------------------------        */
/*         d a s h _ p r o x y _ s e g m e n t          */
/*      ----------------------------------------        */
private int dash_proxy_segment(request_rec *r)
{
        int     status;
        int     exists;
        apr_finfo_t finfo;
        apr_file_t *file;
        char * target=r->filename;
        ap_set_content_type(r, "video/MP2T");
        if ((status = apr_stat(&finfo, target, APR_FINFO_MIN, r->pool)) != APR_SUCCESS)
                return( dash_cache_segment( target, r ) );
        else if (!( exists = ((finfo.filetype != APR_NOFILE) &&  !(finfo.filetype & APR_DIR)) ))
                return( dash_cache_segment( target, r ) );
        else    return( dash_cached_object( target, r ) );
}

#ifdef	_DEBUG_DASH_CHUNKS
/*      ----------------------------------------        */
/*         d a s h _ p r o x y _ d o c u m e n t        */
/*      ----------------------------------------        */
private int dash_proxy_document(request_rec *r)
{
	return( dash_cached_object( r->filename, r ) );
}
#endif

#ifdef	_DASH_INFO
/*	-------------------------------		*/
/*	d a s h _ p r o x y _ i n f o s 	*/
/*	-------------------------------		*/
private int dash_proxy_infos(request_rec *r)
{
        ap_set_content_type(r, "text/html");
	ap_rprintf(r,"<html><head><title>DASH PROXY Configuration</title></head>\n");
	ap_rprintf(r,"<body><div align=center><table border=1><tr><th colspan=2>DASH PROXY Configuration</th></tr>\n");
	ap_rprintf(r,"<tr><th>HOST<th>%s</tr>\n",DashProxy.host);
	ap_rprintf(r,"<tr><th>PORT<th>%u</tr>\n",DashProxy.port);
	ap_rprintf(r,"<tr><th>TTL<th>%u</tr>\n",DashProxy.ttl);
	ap_rprintf(r,"<tr><th>PATH<th>%s</tr>\n",DashProxy.path);
	ap_rprintf(r,"<tr><th>NAME<th>%s</tr>\n",DashProxy.name);
	ap_rprintf(r,"<tr><th>USER<th>%s</tr>\n",DashProxy.user);
	ap_rprintf(r,"<tr><th>PASS<th>%s</tr>\n",DashProxy.pass);
	ap_rprintf(r,"<tr><th>AUTH<th>%s</tr>\n",DashProxy.auth);
	ap_rprintf(r,"<tr><th>AGENT<th>%s</tr>\n",DashProxy.agent);
	ap_rprintf(r,"<tr><th>RETRY<th>%u</tr>\n",DashProxy.retry);
	ap_rprintf(r,"<tr><th>RETRIES<th>%u</tr>\n",DashProxy.retries);
	ap_rprintf(r,"</table></div></body></html>\n");
        return( OK );
}
#endif

/*	------------------	*/
/*	dash_proxy_handler	*/
/*	------------------	*/
private	int dash_proxy_handler(request_rec *r)
{
	char * filename;
	char * filetype;

	/* -------------------------------------- */
	/* check that the module has been enabled */
	/* -------------------------------------- */
	if (!( DashProxy.enabled ))
	{
		dash_log_message(r,0,"DASH proxy declined","not enabled");
		return(DECLINED);
	}

	/* ----------------------------------- */
	/* check the correct handler directive */
	/* ----------------------------------- */
	else if ((!(r->handler)) || ( strcmp(r->handler, "dash-proxy-handler")) )
	{
		dash_log_message(r,0,"DASH proxy declined","incorrect handler");
		return (DECLINED);
	}

	/* ---------------------------------- */
	/* we only accept GET method requests */
	/* ---------------------------------- */
    	else if ( strcasecmp(r->method, "GET") != 0)
	{
		dash_log_message(r,INCORRECT_METHOD,"incorrect method",r->method);
		return( INCORRECT_METHOD );
	}
	/* ------------------------------ */
	/* check for matching path prefix */
	/* ------------------------------ */
    	else if ( strncmp(r->filename, DashProxy.path, strlen( DashProxy.path )) != 0)
	{
		dash_log_message(r,INCORRECT_METHOD,"incorrect path",r->method);
		return( INCORRECT_METHOD );
	}
	else
	{
		/* -------------------------- */
		/* separate filename and type */
		/* -------------------------- */
		filename = apr_pstrdup(r->pool, r->filename);
		for ( 	filetype = filename;
			*filetype != 0;
			filetype++)
		{
			if ( *filetype =='.' )
			{
				*(filetype++) = 0;
				break;
			}
		}
		/* ------------------------ */
		/* is it a manifest request */
		/* ------------------------ */
		if (!( strcasecmp( filetype,"mpd" ) ))
			return( dash_proxy_manifest( r ) );

		/* ------------------------ */
		/* is it a segment request  */
		/* ------------------------ */
		else if (!( strcasecmp( filetype,"m4s" ) ))
			return( dash_proxy_segment( r ) );

#ifdef	_DEBUG_DASH_CHUNKS
                /* --------------------- */
                /* is it an test request */
                /* --------------------- */
                else if (!( strcasecmp( filetype,"dash" ) ))
                        return( dash_proxy_document( r ) );
#endif

#ifdef	_DASH_INFO
		/* --------------------- */
		/* DASH PROXY INFORMATION */
		/* --------------------- */
                else if (!( strcasecmp( filetype,"info" ) ))
                        return( dash_proxy_infos( r ) );
#endif

		else
		{
			dash_log_message(r, HTTP_NOT_FOUND,"incorrect type", filetype );
			return( HTTP_NOT_FOUND ); 
		}
	}
}

/*	-------------------------	*/
/*	dash_proxy_register_hooks	*/
/*	-------------------------	*/
private	void register_hooks(apr_pool_t *pool)
{
    ap_hook_handler(dash_proxy_handler, NULL, NULL, APR_HOOK_LAST);
}

	/* ----------------- */
#endif	/* _mod_dash_proxy_c */
	/* ----------------- */

